from time import time
import unittest
from unittest.mock import patch, Mock
from urllib.error import URLError

from suds import WebFault

from ..exceptions import (
    ConnectError, ServiceError, ApiLimitError, AccountFault, TableFault, ListFault)
from .. import client


class InteractClientTests(unittest.TestCase):
    """ Test InteractClient """

    def setUp(self):
        self.client = Mock()
        self.configuration = {
            'username': 'username',
            'password': 'password',
            'pod': 'pod',
            'client': self.client,
        }
        self.interact = client.InteractClient(**self.configuration)

    def test_starts_disconnected(self):
        self.assertFalse(self.interact.connected)

    @patch.object(client, 'time')
    def test_connected_property_returns_time_of_connection_after_successful_connect(self, mtime):
        mtime.return_value = connection_time = time()
        self.interact.connect()
        self.assertEqual(self.interact.connected, connection_time)

    @patch.object(client, 'time')
    @patch.object(client.InteractClient, 'login')
    def test_session_property_returns_session_id_and_start_after_successful_connect(
            self, login, mtime):

        mtime.return_value = session_start = time()
        session_id = "session_id"
        login.return_value = Mock(session_id=session_id)
        self.interact.connect()

        self.assertEqual(self.interact.session, (session_id, session_start))

    @patch.object(client.InteractClient, 'login')
    def test_connect_reuses_session_if_possible_and_does_not_login(
            self, login):
        self.interact.session = "session_id"

        self.interact.connect()
        self.assertFalse(login.called)

    @patch.object(client.InteractClient, 'login')
    def test_connect_gets_new_session_if_session_is_expired(self, login):
        self.interact.connect()
        self.interact.disconnect()
        self.interact.session_lifetime = -1

        self.interact.connect()
        self.assertEqual(login.call_count, 2)

    def test_connected_property_returns_false_after_disconnect(self):
        self.interact.disconnect()
        self.assertFalse(self.interact.connected)

    def test_client_property_returns_configured_client(self):
        self.assertEqual(self.interact.client, self.client)

    def test_call_method_calls_soap_method_with_passed_arguments(self):
        self.interact.call('somemethod', 'arg')
        self.client.service.somemethod.assert_called_with('arg')

    def test_call_method_returns_soap_method_return_value(self):
        self.client.service.bananas.return_value = 1
        self.assertEqual(self.interact.call('bananas'), 1)

    def test_call_method_raises_ConnectError_for_url_timeout(self):
        self.client.service.rm_rf.side_effect = URLError('Timeout')
        with self.assertRaises(ConnectError):
            self.interact.call('rm_rf', '/.')

    def test_call_method_raises_ServiceError_for_unhandled_webfault(self):
        self.client.service.rm_rf.side_effect = WebFault(Mock(), Mock())
        with self.assertRaises(ServiceError):
            self.interact.call('rm_rf', '/.')

    def test_call_method_raises_ListFault_for_list_fault_exception_from_service(self):
        self.client.service.list_thing.side_effect = WebFault(
            Mock(faultstring='ListFault'), Mock())
        with self.assertRaises(ListFault):
            self.interact.call('list_thing')

    def test_call_method_raises_ApiLimitError_for_rate_limit_exception_from_service(self):
        self.client.service.rm_rf.side_effect = WebFault(
            Mock(faultstring='API_LIMIT_EXCEEDED'), Mock())
        with self.assertRaises(ApiLimitError):
            self.interact.call('rm_rf', '/.')

    def test_call_method_raises_TableFault_for_table_fault_exception_from_service(self):
        self.client.service.give_me_a_table.side_effect = WebFault(
            Mock(faultstring='TableFault'), Mock())
        with self.assertRaises(TableFault):
            self.interact.call('give_me_a_table', 'awesome_table')

    @patch.object(client.InteractClient, 'WSDLS', {'pod': 'pod_wsdl'})
    def test_wsdl_property_returns_correct_value(self):
        self.assertEqual(self.interact.wsdl, 'pod_wsdl')

    @patch.object(client.InteractClient, 'ENDPOINTS', {'pod': 'pod_endpoint'})
    def test_endpoint_property_returns_correct_value(self):
        self.assertEqual(self.interact.endpoint, 'pod_endpoint')

    @patch.object(client.InteractClient, 'connect', Mock())
    def test_entering_context_calls_connect(self):
        self.assertFalse(self.interact.connect.called)
        with self.interact:
            self.assertTrue(self.interact.connect.called)

    @patch.object(client.InteractClient, 'disconnect', Mock())
    def test_leaving_context_calls_disconnect(self):
        with self.interact:
            self.assertFalse(self.interact.disconnect.called)
        self.assertTrue(self.interact.disconnect.called)

    @patch.object(client.InteractClient, 'login')
    def test_connect_method_raises_account_fault_on_credential_failure(self, login):
        login.side_effect = AccountFault
        with self.assertRaises(AccountFault):
            self.interact.connect()

    @patch.object(client.InteractClient, 'login', Mock(return_value=Mock(sessionId=1)))
    def test_connect_method_returns_true_on_success(self):
        self.assertTrue(self.interact.connect())

    def test_connect_method_sets_soapheaders(self):
        soapheaders = Mock()
        self.interact.client.factory.create.return_value = soapheaders
        self.interact.connect()
        self.interact.client.set_options.assert_called_once_with(soapheaders=soapheaders)

    @patch.object(client.InteractClient, 'login')
    @patch.object(client.InteractClient, 'logout')
    def test_connect_abandons_session_if_session_is_expired(self, logout, login):
        self.interact.session_lifetime = -1
        self.interact.session = session_id = '1234'

        self.interact.connect()
        logout.assert_called_once_with()
        self.assertNotEqual(self.interact.session[0], session_id)

    @patch.object(client.InteractClient, 'logout')
    def test_disconnect_does_not_logout_if_session_is_available(self, logout):
        self.session = 'session_id'

        self.interact.disconnect()
        self.assertEqual(logout.call_count, 0)

    @patch.object(client.InteractClient, 'logout')
    def test_disconnect_calls_logout_if_session_is_expired(self, logout):
        self.interact.session = 'session_id'
        self.interact.session_lifetime = -1

        self.interact.disconnect()
        self.assertEqual(logout.call_count, 1)
        self.assertIsNone(self.interact.session)

    @patch.object(client.InteractClient, 'logout')
    def test_disconnect_calls_logout_if_abandon_session_is_passed(self, logout):
        self.interact.connect()
        self.interact.disconnect(abandon_session=True)
        self.assertEqual(logout.call_count, 1)
        self.assertIsNone(self.interact.session)




import unittest

from mock import Mock, patch

from ..types import (
    InteractType, InteractObject, ListMergeRule, RecordData, Record, DeleteResult, LoginResult,
    MergeResult, RecipientResult, ServerAuthResult, RecipientData, Recipient, OptionalData)


class InteractTypeTests(unittest.TestCase):

    """ InteractType instance """

    def setUp(self):
        self.type = InteractType(foo='bar')
        self.client = Mock()

    def test_soap_attribute_method_sets_attribute(self):
        """soap_attribute method sets attribute """
        self.type.soap_attribute('red', True)
        self.assertTrue(self.type.red)

    def test_soap_attribute_method_registers_attribute(self):
        """soap_attribute method registers attribute """
        self.type.soap_attribute('blue', True)
        self.assertIn('blue', self.type._attributes)

    def test_soap_name_property_returns_class_name(self):
        """soap_name property returns class name """
        self.assertEqual(self.type.soap_name, self.type.__class__.__name__)

    def test_instance_provides_attributes_through_dictionary_lookup(self):
        self.assertEqual(self.type.foo, self.type['foo'])

    def test_get_soap_object_method_returns_client_factory_type(self):
        """get_soap_object_method returns client factory type """
        soap_object = self.client.factory.create.return_value = Mock()
        self.assertEqual(self.type.get_soap_object(self.client), soap_object)

    def test_get_soap_object_method_returns_object_with_correct_attributes_set(self):
        """get_soap_object method returns object with correct attributes set """
        self.type.soap_attribute('red_fish', True)
        soap_object = self.type.get_soap_object(self.client)
        self.assertTrue(all(
            [hasattr(soap_object, attr) for attr in ['redFish', 'foo']]))


class TypeEqualityTests(unittest.TestCase):
    def setUp(self):
        self.customer_id = 1
        self.folder_name = 'folder'
        self.object_name = 'object'
        self.interact_object = InteractObject(self.folder_name, self.object_name)
        self.optional_data = OptionalData({'one': 1})
        self.recipient = Recipient(self.interact_object, customer_id=self.customer_id)
        self.recipient_data = RecipientData(self.recipient, self.optional_data)

    def test_interact_object_equality(self):
        self.assertEqual(self.interact_object, InteractObject(self.folder_name, self.object_name))
        self.assertNotEqual(self.interact_object, InteractObject('three', 'four'))

    def test_recipient_object_equality(self):
        self.assertEqual(
            self.recipient, Recipient(self.interact_object, customer_id=self.customer_id))
        self.assertNotEqual(self.recipient, Recipient(self.interact_object, customer_id=2))

    def test_recipient_data_object_equality(self):
        self.assertEqual(self.recipient_data, RecipientData(self.recipient, self.optional_data))
        self.assertNotEqual(self.recipient_data, RecipientData(self.recipient, OptionalData({})))


class InteractTypeChildTests(unittest.TestCase):

    """ InteractType descendant """

    @classmethod
    def generate_type_methods(cls, types_and_expectations):
        """ Auto generates test methods for type attribute tests """
        for TypeClass, init, attrs in types_and_expectations:
            def create_test_func(TypeClass, init, attrs):
                def test_method(self):
                    instance = TypeClass(**init)
                    for attr in attrs:
                        self.assertEqual(getattr(instance, attr), attrs[attr])

                test_method.__name__ = 'test_%s_has_expected_attributes' % TypeClass.__name__
                return test_method

            test_method = create_test_func(TypeClass, init, attrs)
            setattr(cls, test_method.__name__, test_method)
        return


InteractTypeChildTests.generate_type_methods([
    # (TypeToTest
    #   initializer kwargs,
    #   attributes expectations)
    (InteractObject,
        {'folder_name': 'blarg', 'object_name': 'fuuuuu'},
        {'folder_name': 'blarg', 'object_name': 'fuuuuu'}),
    (DeleteResult,
        {'delete_result': Mock(errorMessage='', success=True, exceptionCode='', id=1)},
        {'error_message': '', 'success': True, 'exception_code': '', 'id': 1}),
    (LoginResult,
        {'login_result': Mock(sessionId=1)},
        {'session_id': 1}),
    (ListMergeRule,
        {'insert_on_no_match': 'A'},
        {'insert_on_no_match': 'A'}),
    (Record,
        {'record': [1, 2, 3]},
        {'field_values': [1, 2, 3]}),
    (MergeResult,
        {'merge_result': Mock(insertCount=1, updateCount=1, rejectedCount=1, totalCount=3,
         errorMessage='Blarg')},
        {'insert_count': 1, 'update_count': 1, 'rejected_count': 1, 'total_count': 3,
         'error_message': 'Blarg'}),
    (RecipientResult,
        {'recipient_result': Mock(recipientId=1, errorMessage='Blarg')},
        {'recipient_id': 1, 'error_message': 'Blarg'}),
    (ServerAuthResult,
        {'server_auth_result': Mock(authSessionId=1, encryptedClientChallenge='boo',
         serverChallenge='ahhh')},
        {'auth_session_id': 1, 'encrypted_client_challenge': 'boo', 'server_challenge': 'ahhh'})
])


class RecordDataTests(unittest.TestCase):
    def setUp(self):
        self.record_patcher = patch('responsys.types.Record')
        self.Record = self.record_patcher.start()
        self.addCleanup(self.record_patcher.stop)

        self.record_data = RecordData([{'foo': 1, 'bar': 2}, {'bar': 4, 'foo': 3}])

    def test_sets_proper_field_name_values(self):
        self.assertEqual(set(self.record_data.field_names), set(['foo', 'bar']))

    def test_sets_proper_record_values(self):
        self.assertTrue(
            self.record_data.records == [[1, 2], [3, 4]] or
            self.record_data.records == [[2, 1], [4, 3]])


class MergeResultTests(unittest.TestCase):
    def setUp(self):
        self.error_message = 'These failed: Record 1 = Test, Record 2 = What'
        self.merge_result = MergeResult(Mock(
            insertCount=1,
            updateCount=1,
            rejectedCount=2,
            totalCount=4,
            errorMessage=self.error_message,
        ))

    def test_failed_property_returns_list_of_ids_from_error_string(self):
        self.assertEqual(self.merge_result.failed, [1, 2])



import re
from collections import UserDict


class InteractType(object):

    """ InteractType class

    Provides base interact type functionality. Interact types should register their WSDL defined
    attributes via the soap_attribute method. This allows interact types to provide their own
    soap friendly objects for use with the suds client used by the InteractClient.

    Interact type attributes can be accessed via dictionary lookup, for example:

        >>> InteractType(foo=1)
        >>> InteractType['foo'] == InteractType.foo
        ... True
    """

    def __init__(self, *args, **kwargs):
        self._attributes = set()
        self.set_attributes(*args, **kwargs)

    def __getitem__(self, name):
        return getattr(self, name)

    @property
    def soap_name(self):
        """ Provide the WSDL defined name for this class. """
        return self.__class__.__name__

    def soap_attribute(self, name, value):
        """ Marks an attribute as being a part of the data defined by the soap datatype"""
        setattr(self, name, value)
        self._attributes.add(name)

    def get_soap_object(self, client):
        """ Create and return a soap service type defined for this instance """
        def to_soap_attribute(attr):
            words = attr.split('_')
            words = words[:1] + [word.capitalize() for word in words[1:]]
            return ''.join(words)

        soap_object = client.factory.create(self.soap_name)
        for attr in self._attributes:
            value = getattr(self, attr)
            setattr(soap_object, to_soap_attribute(attr), value)

        return soap_object

    def set_attributes(self, *args, **kwargs):
        for name, value in list(kwargs.items()):
            self.soap_attribute(name, value)

    def __eq__(self, a):
        attr_equal = lambda attr: getattr(self, attr) == getattr(a, attr)
        return all([attr_equal(attr) for attr in self._attributes])


class InteractObject(InteractType):

    """ Responsys InteractObject Type """

    def set_attributes(self, folder_name, object_name):
        self.soap_attribute('folder_name', folder_name)
        self.soap_attribute('object_name', object_name)


class ListMergeRule(InteractType):

    """ ListMergeRule

    Constructor accepts overrides for the following defaults:

        'insert_on_no_match': True,
        'update_on_match': 'REPLACE_ALL',
        'match_column_name_1': 'Customer_Id_',
        'match_column_name_2': None,
        'match_column_name_3': None,
        'match_operator': 'NONE',
        'optin_value': 'I',
        'optout_value': 'O',
        'html_value': 'H',
        'text_value': 'T',
        'reject_record_if_channel_empty': 'E',
        'default_permission_status': 'OPTIN',

    See Responsys API documentation for more information on what values are available and what
    these options do.
    """

    DEFAULTS = {
        'insert_on_no_match': True,
        'update_on_match': 'REPLACE_ALL',
        'match_column_name_1': 'Customer_Id_',
        'match_column_name_2': None,
        'match_column_name_3': None,
        'match_operator': 'NONE',
        'optin_value': 'I',
        'optout_value': 'O',
        'html_value': 'H',
        'text_value': 'T',
        'reject_record_if_channel_empty': 'E',
        'default_permission_status': 'OPTIN',
    }

    def set_attributes(self, **overrides):
        options = self.DEFAULTS.copy()
        options.update(overrides)

        self.soap_attribute('insert_on_no_match', options['insert_on_no_match'])
        self.soap_attribute('update_on_match', options['update_on_match'])
        self.soap_attribute('match_column_name_1', options['match_column_name_1'])
        self.soap_attribute('match_column_name_2', options['match_column_name_2'])
        self.soap_attribute('match_column_name_3', options['match_column_name_3'])
        self.soap_attribute('match_operator', options['match_operator'])
        self.soap_attribute('optin_value', options['optin_value'])
        self.soap_attribute('optout_value', options['optout_value'])
        self.soap_attribute('html_value', options['html_value'])
        self.soap_attribute('text_value', options['text_value'])
        self.soap_attribute(
            'reject_record_if_channel_empty', options['reject_record_if_channel_empty'])
        self.soap_attribute('default_permission_status', options['default_permission_status'])


class RecordData(InteractType):

    """ Responsys RecordData Type

    Responsys type representing a mapping of field names to values. Accepts a list of dictionary
    like objects for init.
    """

    @classmethod
    def from_soap_type(cls, record_data):
        record_data = [
            dict(zip(record_data.fieldNames, r.fieldValues)) for r in record_data.records]
        return cls(record_data)

    def set_attributes(self, record_data):
        assert len(record_data), "Record list length must be non-zero"
        field_names = list(record_data[0].keys())

        records = []
        for record in record_data:
            records.append([record[field_name] for field_name in field_names])

        self.soap_attribute('field_names', field_names)
        self.soap_attribute('records', records)

    def __iter__(self):
        for record in self.records:
            yield dict(zip(self.field_names, record.field_values))

    def __len__(self):
        return len(self.records)

    def get_soap_object(self, client):
        """ Override default get_soap_object behavior to account for child Record types """
        record_data = super().get_soap_object(client)
        record_data.records = [Record(r).get_soap_object(client) for r in record_data.records]
        return record_data


class Record(InteractType):

    """ Responsys Record Type

    A record is a series of values. Can be iterated over and has a length.
    """

    def set_attributes(self, record):
        field_values = list(record)
        self.soap_attribute('field_values', field_values)

    def __iter__(self):
        return (v for v in self.field_values)

    def __len__(self):
        return len(self.field_values)


class DeleteResult(InteractType):
    """ Responsys DeleteResult Type """
    def set_attributes(self, delete_result):
        self.soap_attribute('error_message', delete_result.errorMessage)
        self.soap_attribute('success', delete_result.success)
        self.soap_attribute('exception_code', delete_result.exceptionCode)
        self.soap_attribute('id', delete_result.id)


class LoginResult(InteractType):
    def set_attributes(self, login_result):
        self.soap_attribute('session_id', login_result.sessionId)


class MergeResult(InteractType):
    def set_attributes(self, merge_result):
        self.soap_attribute('insert_count', merge_result.insertCount)
        self.soap_attribute('update_count', merge_result.updateCount)
        self.soap_attribute('rejected_count', merge_result.rejectedCount)
        self.soap_attribute('total_count', merge_result.totalCount)
        self.soap_attribute('error_message', merge_result.errorMessage)

    @property
    def failed(self):
        failed = None
        if self.error_message:
            failed = re.findall(r'Record ([0-9]*) =', self.error_message)
            failed = [f.isnumeric() and int(f) or f for f in failed]

        return failed or []


class RecipientResult(InteractType):
    def set_attributes(self, recipient_result):
        self.soap_attribute('recipient_id', recipient_result.recipientId)
        self.soap_attribute('error_message', recipient_result.errorMessage)


class ServerAuthResult(InteractType):
    def set_attributes(self, server_auth_result):
        self.soap_attribute('auth_session_id', server_auth_result.authSessionId)
        self.soap_attribute(
            'encrypted_client_challenge', server_auth_result.encryptedClientChallenge)
        self.soap_attribute('server_challenge', server_auth_result.serverChallenge)


class CustomEvent(InteractType):
    def get_soap_object(self, client):
        custom_event = client.factory.create(self.soap_name)
        custom_event.eventName = self.event_name
        custom_event.eventId = self.event_id
        custom_event.eventStringDataMapping = self.event_string_data_mapping
        custom_event.eventDateDataMapping = self.event_date_data_mapping
        custom_event.eventNumberDataMapping = self.event_number_data_mapping
        return custom_event

    def set_attributes(self, event_name, event_id, event_string_data_mapping=None,
            event_date_data_mapping=None, event_number_data_mapping=None):
        self.soap_attribute('event_name', event_name)
        self.soap_attribute('event_id', event_id)
        self.soap_attribute('event_string_data_mapping', event_string_data_mapping)
        self.soap_attribute('event_date_data_mapping', event_date_data_mapping)
        self.soap_attribute('event_number_data_mapping', event_number_data_mapping)


class Recipient(InteractType):
    class EmailFormats(object):
        TEXT = 'TEXT_FORMAT'
        HTML = 'HTML_FORMAT'
        MULTIPART = 'MULTIPART_FORMAT'
        NONE = 'NO_FORMAT'

    def get_soap_object(self, client):
        recipient = client.factory.create(self.soap_name)
        recipient.listName = self.list_name.get_soap_object(client)
        recipient.recipientId = self.recipient_id
        recipient.customerId = self.customer_id
        recipient.emailAddress = self.email_address
        recipient.mobileNumber = self.mobile_number
        recipient.emailFormat = self.email_format
        return recipient

    def set_attributes(
            self, list_name, recipient_id=None, customer_id=None,
            email_address=None, mobile_number=None, email_format=EmailFormats.TEXT):

        assert any([recipient_id, customer_id, email_address, mobile_number]), (
            "At least one of recipient_id, customer_id, mobile_number, or email_address must be "
            "provided")
        self.soap_attribute('list_name', list_name)
        self.soap_attribute('recipient_id', recipient_id)
        self.soap_attribute('customer_id', customer_id)
        self.soap_attribute('email_address', email_address)
        self.soap_attribute('mobile_number', mobile_number)
        self.soap_attribute('email_format', email_format)


class RecipientData(InteractType):
    def set_attributes(self, recipient, optional_data=None):
        self.soap_attribute('recipient', recipient)
        self.soap_attribute('optional_data', OptionalData(optional_data) or OptionalData({}))

    def get_soap_object(self, client):
        recipient_data = client.factory.create(self.soap_name)
        recipient_data.optionalData = self.optional_data.get_soap_object(client)
        recipient_data.recipient = self.recipient.get_soap_object(client)
        return recipient_data


class OptionalData(UserDict, InteractType):
    def get_soap_object(self, client):
        optional_data_list = []
        for name, value in self.items():
            optional_data = client.factory.create(self.soap_name)
            optional_data.name = name
            optional_data.value = value
            optional_data_list.append(optional_data)
        return optional_data_list or None


class TriggerResult(InteractType):
    def set_attributes(self, trigger_result):
        self.soap_attribute('recipient_id', trigger_result.recipientId)
        self.soap_attribute('success', trigger_result.success)
        self.soap_attribute('error_message', trigger_result.errorMessage)



import logging
from time import time
from urllib.error import URLError

from suds.client import Client
from suds import WebFault

from .exceptions import (
    ConnectError, ServiceError, AccountFault, ApiLimitError, TableFault, ListFault)
from .types import (
    RecordData, RecipientResult, MergeResult, DeleteResult, LoginResult, ServerAuthResult,
    TriggerResult)

log = logging.getLogger(__name__)


class InteractClient(object):

    """ Interact Client Class

    Provides access to the methods defined by the Responsys Interact API. Example setup:

        >>> client = InteractClient(username, password, pod)
        >>> client.connect()
        >>> client.merge_list_members(interact_object, records, merge_rules)
        >>> client.disconnect()

    Using the client class as a context manager will automatically connect using the credentials
    provided, and disconnect upon context exit:

        >>> with InteractClient(username, password, pod) as client:
        ...     client.merge_list_members(interact_object, records, merge_rules)

    Since responsys limits the number of active sessions per account, this can help ensure you
    don't leave unused connections open.
    """
    DEFAULT_SESSION_LIFETIME = 60 * 10
    WSDLS = {
        '2': 'https://ws2.responsys.net/webservices/wsdl/ResponsysWS_Level1.wsdl',
        '5': 'https://ws5.responsys.net/webservices/wsdl/ResponsysWS_Level1.wsdl',
        'rtm4': 'https://rtm4.responsys.net/tmws/services/TriggeredMessageWS?wsdl',
        'rtm4b': 'https://rtm4b.responsys.net/tmws/services/TriggeredMessageWS?wsdl',
    }

    ENDPOINTS = {
        '2': 'https://ws2.responsys.net/webservices/services/ResponsysWSService',
        '5': 'https://ws5.responsys.net/webservices/services/ResponsysWSService',
        'rtm4': 'http://rtm4.responsys.net:80/tmws/services/TriggeredMessageWS',
        'rtm4b': 'http://rtm4b.responsys.net:80/tmws/services/TriggeredMessageWS',
    }

    @property
    def wsdl(self):
        return self.WSDLS[self.pod]

    @property
    def endpoint(self):
        return self.ENDPOINTS[self.pod]

    @property
    def client(self):
        if self._client is None:
            self._client = Client(self.wsdl, location=self.endpoint, timeout=self.timeout)
        return self._client

    @property
    def connected(self):
        return getattr(self, '_connected', False)

    @connected.setter
    def connected(self, value):
        self._connected = value

    @property
    def session(self):
        return getattr(self, '_session', None)

    @session.setter
    def session(self, session_id):
        self._session = type(
            'Session', (tuple,), {
                'is_expired': property(lambda s: s[1] + self.session_lifetime <= time()),
        })([session_id, time()])

        session_header = self.client.factory.create('SessionHeader')
        session_header.sessionId = session_id
        self.client.set_options(soapheaders=session_header)

    @session.deleter
    def session(self):
        self._session = None
        self.client.set_options(soapheaders=())

    def __init__(self, username, password, pod, client=None, session_lifetime=600, timeout=5):
        self.username = username
        self.password = password
        self.pod = pod
        self.session_lifetime = session_lifetime
        self.timeout = timeout
        self._client = client

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type_, value, traceback):
        self.disconnect()

    def call(self, method, *args):
        """ Calls the service method defined with the arguments provided """
        try:
            response = getattr(self.client.service, method)(*args)
        except URLError:
            raise ConnectError("Request to service timed out")
        except WebFault as web_fault:
            fault_name = getattr(web_fault.fault, 'faultstring', None)
            error = str(web_fault.fault.detail)

            if fault_name == 'TableFault':
                raise TableFault(error)
            if fault_name == 'ListFault':
                raise ListFault(error)
            if fault_name == 'API_LIMIT_EXCEEDED':
                raise ApiLimitError(error)
            if fault_name == 'AccountFault':
                raise AccountFault(error)

            raise ServiceError(web_fault.fault, web_fault.document)
        return response

    def connect(self):
        """ Connects to the Responsys soap service

        Uses the credentials passed to the client init to login and setup the session id returned.
        Returns True on successful connection, otherwise False.
        """

        if self.session and self.session.is_expired:
            # Close the session to avoid max concurrent session errors
            self.disconnect(abandon_session=True)

        if not self.session:

            try:
                login_result = self.login(self.username, self.password)
            except AccountFault:
                log.error('Login failed, invalid username or password')
                raise
            else:
                self.session = login_result.session_id

        self.connected = time()
        return self.connected

    def disconnect(self, abandon_session=False):
        """ Disconnects from the Responsys soap service

        Calls the service logout method and destroys the client's session information. Returns
        True on success, False otherwise.
        """
        self.connected = False
        if (self.session and self.session.is_expired) or abandon_session:
            try:
                self.logout()
            except:
                log.warning(
                    'Logout call to responsys failed, session may have not been terminated',
                    exc_info=True
                )
            del self.session
        return True

    # Session Management Methods
    def login(self, username, password):
        """ Responsys.login soap call

        Accepts username and password for authentication, returns a LoginResult object.
        """
        return LoginResult(self.call('login', username, password))

    def logout(self):
        """ Responsys.logout soap call

        Returns True on success, False otherwise.
        """
        return self.call('logout')

    def login_with_certificate(self, encrypted_server_challenge):
        """ Responsys.loginWithCertificate soap call

        Accepts encrypted_server_challenge for login. Returns LoginResult.
        """
        return LoginResult(self.call('loginWithCertificate', encrypted_server_challenge))

    def authenticate_server(self, username, client_challenge):
        """ Responsys.authenticateServer soap call

        Accepts username and client_challenge to authenciate. Returns ServerAuthResult.
        """
        return ServerAuthResult(self.call('authenticateServer', username, client_challenge))

    # List Management Methods
    def merge_list_members(self, list_, record_data, merge_rule):
        """ Responsys.mergeListMembers call

        Accepts:
            InteractObject list_
            RecordData record_data
            ListMergeRule merge_rule

        Returns a MergeResult
        """
        list_ = list_.get_soap_object(self.client)
        record_data = record_data.get_soap_object(self.client)
        merge_rule = merge_rule.get_soap_object(self.client)
        return MergeResult(self.call('mergeListMembers', list_, record_data, merge_rule))

    def merge_list_members_RIID(self, list_, record_data, merge_rule):
        """ Responsys.mergeListMembersRIID call

        Accepts:
            InteractObject list_
            RecordData record_data
            ListMergeRule merge_rule

        Returns a RecipientResult
        """
        list_ = list_.get_soap_object(self.client)
        result = self.call('mergeListMembersRIID', list_, record_data, merge_rule)
        return RecipientResult(result.recipientResult)

    def delete_list_members(self, list_, query_column, ids_to_delete):
        """ Responsys.deleteListMembers call

        Accepts:
            InteractObject list_
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list ids_to_delete

        Returns a list of DeleteResult instances
        """
        list_ = list_.get_soap_object(self.client)
        result = self.call('deleteListMembers', list_, query_column, ids_to_delete)
        if hasattr(result, '__iter__'):
            return [DeleteResult(delete_result) for delete_result in result]
        return [DeleteResult(result)]

    def retrieve_list_members(self, list_, query_column, field_list, ids_to_retrieve):
        """ Responsys.retrieveListMembers call

        Accepts:
            InteractObject list_
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list field_list
            list ids_to_retrieve

        Returns a RecordData instance
        """
        list_ = list_.get_soap_object(self.client)
        result = self.call('retrieveListMembers', list_, query_column, field_list, ids_to_retrieve)
        return RecordData.from_soap_type(result.recordData)

    # Table Management Methods
    def create_table(self, table, fields):
        """ Responsys.createTable call

        Accepts:
            InteractObject table
            list fields

        Returns True on success
        """
        table = table.get_soap_object(self.client)
        return self.call('createTable', table, fields)

    def create_table_with_pk(self, table, fields, primary_keys):
        """ Responsys.createTableWithPK call

        Accepts:
            InteractObject table
            list fields
            list primary_keys

        Returns True on success
        """
        table = table.get_soap_object(self.client)
        return self.call('createTableWithPK', table, fields, primary_keys)

    def delete_table(self, table):
        """ Responsys.deleteTable call

        Accepts:
            InteractObject table

        Returns True on success
        """
        table = table.get_soap_object(self.client)
        return self.call('deleteTable', table)

    def delete_profile_extension_members(self, profile_extension, query_column, ids_to_delete):
        """ Responsys.deleteProfileExtensionRecords call

        Accepts:
            InteractObject profile_extension
            list field_list
            list ids_to_retrieve
            string query_column
                default: 'RIID'

        Returns list of DeleteResults
        """
        profile_extension = profile_extension.get_soap_object(self.client)
        result = self.call(
            'deleteProfileExtensionMembers', profile_extension, query_column, ids_to_delete)
        if hasattr(result, '__iter__'):
            return [DeleteResult(delete_result) for delete_result in result]
        return [DeleteResult(result)]

    def retrieve_profile_extension_records(self, profile_extension, field_list, ids_to_retrieve,
                                           query_column='RIID'):
        """ Responsys.retrieveProfileExtensionRecords call

        Accepts:
            InteractObject profile_extension
            list field_list
            list ids_to_retrieve
            string query_column
                default: 'RIID'

        Returns RecordData
        """
        profile_extension = profile_extension.get_soap_object(self.client)
        return RecordData.from_soap_type(
            self.call('retrieveProfileExtensionRecords',
                      profile_extension, query_column, field_list, ids_to_retrieve))

    def truncate_table(self, table):
        """ Responsys.truncateTable call

        Accepts:
            InteractObject table

        Returns True on success
        """
        table = table.get_soap_object(self.client)
        return self.call('truncateTable', table)

    def delete_table_records(self, table, query_column, ids_to_delete):
        """ Responsys.deleteTableRecords call

        Accepts:
            InteractObject table
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list ids_to_delete

        Returns a list of DeleteResult instances
        """
        table = table.get_soap_object(self.client)
        result = self.call('deleteTableRecords', table, query_column, ids_to_delete)
        if hasattr(result, '__iter__'):
            return [DeleteResult(delete_result) for delete_result in result]
        return [DeleteResult(result)]

    def merge_table_records(self, table, record_data, match_column_names):
        """ Responsys.mergeTableRecords call

        Accepts:
            InteractObject table
            RecordData record_data
            list match_column_names

        Returns a MergeResult
        """
        table = table.get_soap_object(self.client)
        record_data = record_data.get_soap_object(self.client)
        return MergeResult(self.call(
            'mergeTableRecords', table, record_data, match_column_names))

    def merge_table_records_with_pk(self, table, record_data, insert_on_no_match, update_on_match):
        """ Responsys.mergeTableRecordsWithPK call

        Accepts:
            InteractObject table
            RecordData record_data
            string insert_on_no_match
            string update_on_match

        Returns a MergeResult
        """
        table = table.get_soap_object(self.client)
        record_data = record_data.get_soap_object(self.client)
        return MergeResult(self.call(
            'mergeTableRecordsWithPK', table, record_data, insert_on_no_match, update_on_match))

    def merge_into_profile_extension(self, profile_extension, record_data, match_column,
                                     insert_on_no_match, update_on_match):
        """ Responsys.mergeIntoProfileExtension call

        Accepts:
            InteractObject profile_extension
            RecordData record_data
            string match_column
            string insert_on_no_match
            string update_on_match

        Returns a RecipientResult
        """
        profile_extension = profile_extension.get_soap_object(self.client)
        record_data = record_data.get_soap_object(self.client)
        results = self.call(
            'mergeIntoProfileExtension', profile_extension, record_data, match_column,
            insert_on_no_match, update_on_match)
        return [RecipientResult(result) for result in results]

    def retrieve_table_records(self, table, query_column, field_list, ids_to_retrieve):
        """ Responsys.retrieveTableRecords call

        Accepts:
            InteractObject table
            string query_column
                possible values: 'RIID'|'EMAIL_ADDRESS'|'CUSTOMER_ID'|'MOBILE_NUMBER'
            list field_list
            list ids_to_retrieve

        Returns a RecordData
        """
        table = table.get_soap_object(self.client)
        return RecordData.from_soap_type(self.call(
            'retrieveTableRecords', table, query_column, field_list, ids_to_retrieve))

    # Campaign Management Methods
    # TODO: implement
    # GetLaunchStatus
    # LaunchCampaign
    # MergeTriggerEmail
    # ScheduleCampaignLaunch
    # TriggerCampaignMessage
    def trigger_custom_event(self, custom_event, recipient_data=None):
        custom_event = custom_event.get_soap_object(self.client)
        recipient_data = [rdata.get_soap_object(self.client) for rdata in recipient_data]
        results = self.call('triggerCustomEvent', custom_event, recipient_data)
        return [TriggerResult(result) for result in results]


    # TODO: Implement
    #
    # Content Management Methods
    # Folder Management Methods



from suds import WebFault


class AccountFault(RuntimeError):
    pass


class ConnectError(RuntimeError):
    pass


class ApiLimitError(RuntimeError):
    pass


class ListFault(RuntimeError):
    pass


class TableFault(RuntimeError):
    pass


class ServiceError(WebFault):
    pass
